#!/usr/bin/env python

from .regressors import ConformalRidgeRegressor, ConformalNearestNeighboursRegressor
from .classifiers import ConformalNearestNeighboursClassifier 
from .martingale import PluginMartingale